@@ -45,7 +45,19 @@ if sys.platform == "win32":
 from meadowrun.run_job import (
     AllocCloudInstance,
     AllocCloudInstances,
+    AwsSecret,
+    AzureSecret,
+    CondaEnvironmentYmlFile,
+    ContainerAtDigestInterpreter,
+    ContainerInterpreter,
+    ContainerInterpreterBase,
     Deployment,
+    InterpreterSpecFile,
+    LocalCondaInterpreter,
+    LocalInterpreter,
+    LocalPipInterpreter,
+    PipRequirementsFile,
+    PoetryProjectPath,
     run_command,
     run_function,
     run_map,
@@ -53,37 +65,26 @@ from meadowrun.run_job import (
 from meadowrun.run_job_local import LocalHost
 from meadowrun.run_job_core import SshHost
 
-from meadowrun.meadowrun_pb2 import (
-    AwsSecret,
-    AzureSecret,
-    ContainerAtDigest,
-    ContainerAtTag,
-    GitRepoBranch,
-    GitRepoCommit,
-    ServerAvailableContainer,
-    ServerAvailableFile,
-    ServerAvailableFolder,
-    ServerAvailableInterpreter,
-)
-
 
 __all__ = [
     "AllocCloudInstance",
     "AllocCloudInstances",
+    "AwsSecret",
+    "AzureSecret",
+    "CondaEnvironmentYmlFile",
+    "ContainerAtDigestInterpreter",
+    "ContainerInterpreter",
+    "ContainerInterpreterBase",
     "Deployment",
+    "InterpreterSpecFile",
+    "LocalCondaInterpreter",
+    "LocalInterpreter",
+    "LocalPipInterpreter",
+    "PipRequirementsFile",
+    "PoetryProjectPath",
     "run_command",
     "run_function",
     "run_map",
     "LocalHost",
     "SshHost",
-    "AwsSecret",
-    "AzureSecret",
-    "ContainerAtDigest",
-    "ContainerAtTag",
-    "GitRepoBranch",
-    "GitRepoCommit",
-    "ServerAvailableContainer",
-    "ServerAvailableFile",
-    "ServerAvailableFolder",
-    "ServerAvailableInterpreter",
 ]
