@@ -8,6 +8,7 @@ import pickle
 import traceback
 import uuid
 from typing import (
+    TYPE_CHECKING,
     Any,
     Callable,
     Iterable,
@@ -34,6 +35,9 @@ from meadowrun.aws_integration.management_lambdas.ec2_alloc_stub import (
     _EC2_ALLOC_TAG_VALUE,
 )
 from meadowrun.instance_allocation import allocate_jobs_to_instances
+
+if TYPE_CHECKING:
+    from meadowrun.instance_selection import Resources
 from meadowrun.meadowrun_pb2 import GridTask, ProcessState, GridTaskStateResponse
 from meadowrun.run_job_core import RunMapHelper, AllocCloudInstancesInternal
 from meadowrun.shared import pickle_exception
@@ -410,10 +414,9 @@ async def prepare_ec2_run_map(
     function: Callable[[_T], _U],
     tasks: Sequence[_T],
     region_name: Optional[str],
-    logical_cpu_required_per_task: int,
-    memory_gb_required_per_task: float,
-    interruption_probability_threshold: float,
+    resources_required_per_task: Resources,
     num_concurrent_tasks: int,
+    ports: Optional[Sequence[str]],
 ) -> RunMapHelper:
     """This code is tightly coupled with run_map"""
 
@@ -430,12 +433,11 @@ async def prepare_ec2_run_map(
         allocated_hosts = await allocate_jobs_to_instances(
             instance_registrar,
             AllocCloudInstancesInternal(
-                logical_cpu_required_per_task,
-                memory_gb_required_per_task,
-                interruption_probability_threshold,
+                resources_required_per_task,
                 num_concurrent_tasks,
                 region_name,
             ),
+            ports,
         )
 
     request_queue, result_queue = await queues_future
@@ -443,9 +445,10 @@ async def prepare_ec2_run_map(
     return RunMapHelper(
         region_name,
         allocated_hosts,
-        functools.partial(
+        worker_function=functools.partial(
             worker_loop, function, request_queue, result_queue, region_name
         ),
-        {"user": SSH_USER, "connect_kwargs": {"pkey": pkey}},
-        get_results(result_queue, region_name, len(tasks)),
+        ssh_username=SSH_USER,
+        ssh_private_key=pkey,
+        results_future=get_results(result_queue, region_name, len(tasks)),
     )
