@@ -3,9 +3,11 @@ from __future__ import annotations
 import dataclasses
 import datetime
 import json
-from types import TracebackType
-from typing import Tuple, List, Optional, Sequence, Type, Any, Dict
-from typing_extensions import Literal
+from typing import TYPE_CHECKING, Any, Iterable, List, Optional, Sequence, Tuple, Type
+
+if TYPE_CHECKING:
+    from typing_extensions import Literal
+    from types import TracebackType
 
 import meadowrun.azure_integration.azure_vms
 from meadowrun.azure_integration.azure_meadowrun_core import (
@@ -13,6 +15,17 @@ from meadowrun.azure_integration.azure_meadowrun_core import (
     get_default_location,
 )
 from meadowrun.azure_integration.azure_ssh_keys import ensure_meadowrun_key_pair
+from meadowrun.azure_integration.mgmt_functions.azure_constants import (
+    ALLOCATED_TIME,
+    LAST_UPDATE_TIME,
+    NON_CONSUMABLE_RESOURCES,
+    PREVENT_FURTHER_ALLOCATION,
+    RESOURCES_ALLOCATED,
+    RESOURCES_AVAILABLE,
+    RUNNING_JOBS,
+    SINGLE_PARTITION_KEY,
+    VM_NAME,
+)
 from meadowrun.azure_integration.mgmt_functions.azure_core.azure_exceptions import (
     ResourceExistsError,
     ResourceModifiedError,
@@ -24,31 +37,23 @@ from meadowrun.azure_integration.mgmt_functions.azure_core.azure_storage_api imp
     azure_table_api_paged,
     table_key_url,
 )
-from meadowrun.azure_integration.mgmt_functions.azure_constants import (
-    ALLOCATED_TIME,
-    LAST_UPDATE_TIME,
-    LOGICAL_CPU_ALLOCATED,
-    LOGICAL_CPU_AVAILABLE,
-    MEMORY_GB_ALLOCATED,
-    MEMORY_GB_AVAILABLE,
-    RUNNING_JOBS,
-    SINGLE_PARTITION_KEY,
-    VM_NAME,
-)
 from meadowrun.azure_integration.mgmt_functions.vm_adjust import VM_ALLOC_TABLE_NAME
 from meadowrun.instance_allocation import (
-    _InstanceState,
     InstanceRegistrar,
+    _InstanceState,
+    _TInstanceState,
     allocate_jobs_to_instances,
 )
-from meadowrun.instance_selection import Resources, CloudInstance
-from meadowrun.meadowrun_pb2 import Job
+from meadowrun.instance_selection import CloudInstance, Resources
+
+if TYPE_CHECKING:
+    from meadowrun.meadowrun_pb2 import Job
+
 from meadowrun.run_job_core import AllocCloudInstancesInternal, JobCompletion, SshHost
 
 
 @dataclasses.dataclass
 class AzureVMInstanceState(_InstanceState):
-    name: str
     # See https://microsoft.github.io/AzureTipsAndTricks/blog/tip88.html
     etag: str
 
@@ -105,18 +110,20 @@ class AzureInstanceRegistrar(InstanceRegistrar[AzureVMInstanceState]):
                     "PartitionKey": SINGLE_PARTITION_KEY,
                     "RowKey": public_address,
                     VM_NAME: name,
-                    LOGICAL_CPU_AVAILABLE: resources_available.logical_cpu,
-                    MEMORY_GB_AVAILABLE: resources_available.memory_gb,
+                    RESOURCES_AVAILABLE: json.dumps(resources_available.consumable),
+                    NON_CONSUMABLE_RESOURCES: json.dumps(
+                        resources_available.non_consumable
+                    ),
                     RUNNING_JOBS: json.dumps(
                         {
                             job_id: {
-                                LOGICAL_CPU_ALLOCATED: allocated_resources.logical_cpu,
-                                MEMORY_GB_ALLOCATED: allocated_resources.memory_gb,
+                                RESOURCES_ALLOCATED: allocated_resources.consumable,
                                 ALLOCATED_TIME: now,
                             }
                             for job_id, allocated_resources in running_jobs
                         }
                     ),
+                    PREVENT_FURTHER_ALLOCATION: False,
                     LAST_UPDATE_TIME: now,
                 },
             )
@@ -144,13 +151,13 @@ class AzureInstanceRegistrar(InstanceRegistrar[AzureVMInstanceState]):
         return [
             AzureVMInstanceState(
                 item["RowKey"],
+                item[VM_NAME],
                 Resources(
-                    float(item[MEMORY_GB_AVAILABLE]),
-                    int(item[LOGICAL_CPU_AVAILABLE]),
-                    {},
+                    json.loads(item[RESOURCES_AVAILABLE]),
+                    json.loads(item[NON_CONSUMABLE_RESOURCES]),
                 ),
                 json.loads(item[RUNNING_JOBS]),
-                item[VM_NAME],
+                item[PREVENT_FURTHER_ALLOCATION],
                 item["odata.etag"],
             )
             async for page in azure_table_api_paged(
@@ -161,10 +168,11 @@ class AzureInstanceRegistrar(InstanceRegistrar[AzureVMInstanceState]):
                     "$select": ",".join(
                         [
                             "RowKey",
-                            LOGICAL_CPU_AVAILABLE,
-                            MEMORY_GB_AVAILABLE,
+                            RESOURCES_AVAILABLE,
+                            NON_CONSUMABLE_RESOURCES,
                             RUNNING_JOBS,
                             VM_NAME,
+                            PREVENT_FURTHER_ALLOCATION,
                         ]
                     ),
                 },
@@ -196,10 +204,11 @@ class AzureInstanceRegistrar(InstanceRegistrar[AzureVMInstanceState]):
                     "$select": ",".join(
                         [
                             "RowKey",
-                            LOGICAL_CPU_AVAILABLE,
-                            MEMORY_GB_AVAILABLE,
+                            RESOURCES_AVAILABLE,
+                            NON_CONSUMABLE_RESOURCES,
                             RUNNING_JOBS,
                             VM_NAME,
+                            PREVENT_FURTHER_ALLOCATION,
                         ]
                     )
                 },
@@ -209,11 +218,13 @@ class AzureInstanceRegistrar(InstanceRegistrar[AzureVMInstanceState]):
 
         return AzureVMInstanceState(
             item["RowKey"],
+            item[VM_NAME],
             Resources(
-                float(item[MEMORY_GB_AVAILABLE]), int(item[LOGICAL_CPU_AVAILABLE]), {}
+                json.loads(item[RESOURCES_AVAILABLE]),
+                json.loads(item[NON_CONSUMABLE_RESOURCES]),
             ),
             json.loads(item[RUNNING_JOBS]),
-            item[VM_NAME],
+            item[PREVENT_FURTHER_ALLOCATION],
             item["odata.etag"],
         )
 
@@ -240,19 +251,15 @@ class AzureInstanceRegistrar(InstanceRegistrar[AzureVMInstanceState]):
                 return False
 
             new_running_jobs[job_id] = {
-                LOGICAL_CPU_ALLOCATED: resources_allocated_per_job.logical_cpu,
-                MEMORY_GB_ALLOCATED: resources_allocated_per_job.memory_gb,
+                RESOURCES_ALLOCATED: resources_allocated_per_job.consumable,
                 ALLOCATED_TIME: now,
             }
 
-        new_logical_cpu_available = (
-            instance.get_available_resources().logical_cpu
-            - resources_allocated_per_job.logical_cpu * len(new_job_ids)
-        )
-        new_memory_gb_available = (
-            instance.get_available_resources().memory_gb
-            - resources_allocated_per_job.memory_gb * len(new_job_ids)
+        new_resources_available = instance.get_available_resources().subtract(
+            resources_allocated_per_job.multiply(len(new_job_ids))
         )
+        if new_resources_available is None:
+            return False  # this shouldn't really happen
 
         try:
             # https://docs.microsoft.com/en-us/rest/api/storageservices/update-entity2
@@ -264,8 +271,10 @@ class AzureInstanceRegistrar(InstanceRegistrar[AzureVMInstanceState]):
                 ),
                 json_content={
                     VM_NAME: instance.name,
-                    LOGICAL_CPU_AVAILABLE: new_logical_cpu_available,
-                    MEMORY_GB_AVAILABLE: new_memory_gb_available,
+                    RESOURCES_AVAILABLE: json.dumps(new_resources_available.consumable),
+                    NON_CONSUMABLE_RESOURCES: json.dumps(
+                        new_resources_available.non_consumable
+                    ),
                     RUNNING_JOBS: json.dumps(new_running_jobs),
                     LAST_UPDATE_TIME: now,
                 },
@@ -289,11 +298,8 @@ class AzureInstanceRegistrar(InstanceRegistrar[AzureVMInstanceState]):
             return False
 
         job = instance.get_running_jobs()[job_id]
-        new_logical_cpu_available = (
-            instance.get_available_resources().logical_cpu + job[LOGICAL_CPU_ALLOCATED]
-        )
-        new_memory_gb_available = (
-            instance.get_available_resources().memory_gb + job[MEMORY_GB_ALLOCATED]
+        new_resources_available = instance.get_available_resources().add(
+            Resources(job[RESOURCES_ALLOCATED], {})
         )
         new_running_jobs = instance.get_running_jobs().copy()
         del new_running_jobs[job_id]
@@ -309,8 +315,10 @@ class AzureInstanceRegistrar(InstanceRegistrar[AzureVMInstanceState]):
                 ),
                 json_content={
                     VM_NAME: instance.name,
-                    LOGICAL_CPU_AVAILABLE: new_logical_cpu_available,
-                    MEMORY_GB_AVAILABLE: new_memory_gb_available,
+                    RESOURCES_AVAILABLE: json.dumps(new_resources_available.consumable),
+                    NON_CONSUMABLE_RESOURCES: json.dumps(
+                        new_resources_available.non_consumable
+                    ),
                     RUNNING_JOBS: json.dumps(new_running_jobs),
                     LAST_UPDATE_TIME: now,
                 },
@@ -322,14 +330,52 @@ class AzureInstanceRegistrar(InstanceRegistrar[AzureVMInstanceState]):
             # optimistic concurrency check failed
             return False
 
+    async def set_prevent_further_allocation(
+        self, public_address: str, value: bool
+    ) -> bool:
+        if self._storage_account is None:
+            raise ValueError(
+                "Tried to use AzureInstanceRegistrar without calling __aenter__"
+            )
+
+        instance = await self.get_registered_instance(public_address)
+
+        now = datetime.datetime.utcnow().isoformat()
+
+        try:
+            # https://docs.microsoft.com/en-us/rest/api/storageservices/update-entity2
+            await azure_table_api(
+                "PUT",  # replace rather than merge
+                self._storage_account,
+                table_key_url(
+                    VM_ALLOC_TABLE_NAME, SINGLE_PARTITION_KEY, instance.public_address
+                ),
+                json_content={
+                    VM_NAME: instance.name,
+                    RESOURCES_AVAILABLE: json.dumps(
+                        instance.get_available_resources().consumable
+                    ),
+                    NON_CONSUMABLE_RESOURCES: json.dumps(
+                        instance.get_available_resources().non_consumable
+                    ),
+                    RUNNING_JOBS: json.dumps(instance.running_jobs),
+                    PREVENT_FURTHER_ALLOCATION: value,
+                    LAST_UPDATE_TIME: now,
+                },
+                additional_headers={"If-Match": instance.etag},
+            )
+            return True
+        except ResourceModifiedError:
+            # this is how the API indicates that the etag does not match, i.e. the
+            # optimistic concurrency check failed
+            return False
+
     async def launch_instances(
         self, instances_spec: AllocCloudInstancesInternal
     ) -> Sequence[CloudInstance]:
         return await meadowrun.azure_integration.azure_vms.launch_vms(
-            instances_spec.logical_cpu_required_per_task,
-            instances_spec.memory_gb_required_per_task,
+            instances_spec.resources_required_per_task,
             instances_spec.num_concurrent_tasks,
-            instances_spec.interruption_probability_threshold,
             (await ensure_meadowrun_key_pair(instances_spec.region_name))[1],
             instances_spec.region_name,
         )
@@ -338,12 +384,22 @@ class AzureInstanceRegistrar(InstanceRegistrar[AzureVMInstanceState]):
         # TODO currently Azure instances are open to everyone
         pass
 
+    async def open_ports(
+        self,
+        ports: Optional[Sequence[str]],
+        allocated_existing_instances: Iterable[_TInstanceState],
+        allocated_new_instances: Iterable[CloudInstance],
+    ) -> None:
+        if ports:
+            raise NotImplementedError(
+                "Opening ports on Azure is not implemented, please create an issue at "
+                "https://github.com/meadowdata/meadowrun/issues"
+            )
+
 
 async def run_job_azure_vm_instance_registrar(
     job: Job,
-    logical_cpu_required: int,
-    memory_gb_required: float,
-    eviction_rate: float,
+    resources_required: Resources,
     location: Optional[str],
 ) -> JobCompletion[Any]:
     if not location:
@@ -351,18 +407,17 @@ async def run_job_azure_vm_instance_registrar(
     pkey, public_key = await ensure_meadowrun_key_pair(location)
 
     async with AzureInstanceRegistrar(location, "create") as instance_registrar:
+        if job.ports:
+            raise NotImplementedError(
+                "Support for opening ports on Azure is not yet implemented, please "
+                "create an issue at https://github.com/meadowdata/meadowrun/issues"
+            )
         hosts = await allocate_jobs_to_instances(
             instance_registrar,
-            AllocCloudInstancesInternal(
-                logical_cpu_required, memory_gb_required, eviction_rate, 1, location
-            ),
+            AllocCloudInstancesInternal(resources_required, 1, location),
+            job.ports,
         )
 
-    fabric_kwargs: Dict[str, Any] = {
-        "user": "meadowrunuser",
-        "connect_kwargs": {"pkey": pkey},
-    }
-
     if len(hosts) != 1:
         raise ValueError(f"Asked for one host, but got back {len(hosts)}")
     host, job_ids = list(hosts.items())[0]
@@ -373,4 +428,6 @@ async def run_job_azure_vm_instance_registrar(
     # remains mostly an internal concept
     job.job_id = job_ids[0]
 
-    return await SshHost(host, fabric_kwargs, ("AzureVM", location)).run_job(job)
+    return await SshHost(host, "meadowrunuser", pkey, ("AzureVM", location)).run_job(
+        job
+    )
