@@ -2,8 +2,10 @@ from __future__ import annotations
 
 import asyncio
 import uuid
-from typing import Optional, Tuple, Sequence, Awaitable, Any, Dict
-from typing_extensions import Literal
+from typing import TYPE_CHECKING, Optional, Tuple, Sequence, Awaitable, Any, Dict
+
+if TYPE_CHECKING:
+    from typing_extensions import Literal
 
 from meadowrun.azure_integration.azure_meadowrun_core import (
     _ensure_managed_identity,
@@ -42,9 +44,9 @@ _MEADOWRUN_USERNAME = "meadowrunuser"
 # https://docs.microsoft.com/en-us/cli/azure/sig/image-definition?view=azure-cli-latest#az-sig-image-definition-list-community
 _MEADOWRUN_COMMUNITY_IMAGE_ID = (
     "/CommunityGalleries/meadowprodeastus-e8b60fd5-8978-467b-a1b0-5b83cbf5393d/Images/"
-    "meadowrun"
+    "meadowrun-dev"
 )
-_MEADOWRUN_IMAGE_VERSION = "0.1.13"
+_MEADOWRUN_IMAGE_VERSION = "0.1.14"
 
 
 async def _ensure_virtual_network_and_subnet(location: str) -> str:
@@ -298,10 +300,8 @@ async def _provision_vm(
 
 
 async def launch_vms(
-    logical_cpu_required_per_job: int,
-    memory_gb_required_per_job: float,
+    resources_required_per_job: Resources,
     num_jobs: int,
-    eviction_rate: float,
     ssh_public_key_data: str,
     location: Optional[str] = None,
 ) -> Sequence[CloudInstance]:
@@ -311,30 +311,26 @@ async def launch_vms(
         location = get_default_location()
 
     chosen_instance_types = choose_instance_types_for_job(
-        Resources(memory_gb_required_per_job, logical_cpu_required_per_job, {}),
+        resources_required_per_job,
         num_jobs,
-        eviction_rate,
         await get_vm_types(location),
     )
-    if len(chosen_instance_types) < 1:
-        raise ValueError(
-            f"There were no instance types that could be selected for "
-            f"memory={memory_gb_required_per_job}, cpu={logical_cpu_required_per_job}"
-        )
 
     vm_detail_tasks = []
     chosen_instance_types_repeated = []
 
-    # for development, you can replace this with something like
-    # (
+    image_id = f"{_MEADOWRUN_COMMUNITY_IMAGE_ID}/versions/{_MEADOWRUN_IMAGE_VERSION}"
+    # for development, you can uncomment the following, and replace "Community" with
+    # "Regular" below
+    # image_id = (
     #     "/subscriptions/d740513f-4172-4792-bd29-5194e79d5881/resourceGroups/"
     #     "meadowrun-dev/providers/Microsoft.Compute/galleries/meadowrun.dev.gallery/"
     #     f"images/meadowrun/versions/{_MEADOWRUN_IMAGE_VERSION}"
     # )
-    # and replace "Community" with "Regular" below
-    image_id = f"{_MEADOWRUN_COMMUNITY_IMAGE_ID}/versions/{_MEADOWRUN_IMAGE_VERSION}"
 
+    at_least_one_chosen_instance_type = False
     for instance_type in chosen_instance_types:
+        at_least_one_chosen_instance_type = True
         for _ in range(instance_type.num_instances):
             vm_detail_tasks.append(
                 _provision_vm(
@@ -348,6 +344,12 @@ async def launch_vms(
             )
             chosen_instance_types_repeated.append(instance_type)
 
+    if not at_least_one_chosen_instance_type:
+        raise ValueError(
+            "There were no instance types that could be selected for "
+            f"{resources_required_per_job}"
+        )
+
     vm_details = await asyncio.gather(*vm_detail_tasks)
 
     return [
