@@ -1,4 +1,5 @@
 """See grid_tasks_sqs.py"""
+from __future__ import annotations
 
 import asyncio
 import dataclasses
@@ -10,6 +11,7 @@ import time
 import traceback
 import uuid
 from typing import (
+    TYPE_CHECKING,
     Any,
     Callable,
     Iterable,
@@ -44,6 +46,9 @@ from meadowrun.azure_integration.mgmt_functions.azure_constants import (
     _RESULT_QUEUE_NAME_PREFIX,
 )
 from meadowrun.instance_allocation import allocate_jobs_to_instances
+
+if TYPE_CHECKING:
+    from meadowrun.instance_selection import Resources
 from meadowrun.meadowrun_pb2 import GridTask, GridTaskStateResponse, ProcessState
 from meadowrun.run_job_core import RunMapHelper, AllocCloudInstancesInternal
 from meadowrun.shared import pickle_exception
@@ -289,10 +294,9 @@ async def prepare_azure_vm_run_map(
     function: Callable[[_T], _U],
     tasks: Sequence[_T],
     location: Optional[str],
-    logical_cpu_required_per_task: int,
-    memory_gb_required_per_task: float,
-    interruption_probability_threshold: float,
+    resources_required_per_task: Resources,
     num_concurrent_tasks: int,
+    ports: Optional[Sequence[str]],
 ) -> RunMapHelper:
     """This code is tightly coupled with run_map"""
     if not location:
@@ -302,15 +306,17 @@ async def prepare_azure_vm_run_map(
     queues_future = asyncio.create_task(create_queues_and_add_tasks(location, tasks))
 
     async with AzureInstanceRegistrar(location, "create") as instance_registrar:
+        if ports:
+            raise NotImplementedError(
+                "Support for opening ports on Azure is not yet implemented, please "
+                "create an issue at https://github.com/meadowdata/meadowrun/issues"
+            )
         allocated_hosts = await allocate_jobs_to_instances(
             instance_registrar,
             AllocCloudInstancesInternal(
-                logical_cpu_required_per_task,
-                memory_gb_required_per_task,
-                interruption_probability_threshold,
-                num_concurrent_tasks,
-                location,
+                resources_required_per_task, num_concurrent_tasks, location
             ),
+            ports,
         )
 
     private_key, public_key = await key_pair_future
@@ -319,10 +325,10 @@ async def prepare_azure_vm_run_map(
     return RunMapHelper(
         location,
         allocated_hosts,
-        functools.partial(worker_loop, function, request_queue, result_queue),
-        {
-            "user": "meadowrunuser",
-            "connect_kwargs": {"pkey": private_key},
-        },
-        get_results(result_queue, len(tasks), location),
+        worker_function=functools.partial(
+            worker_loop, function, request_queue, result_queue
+        ),
+        ssh_username="meadowrunuser",
+        ssh_private_key=private_key,
+        results_future=get_results(result_queue, len(tasks), location),
     )
