@@ -1,5 +1,7 @@
 # TODO turn this into a real module that allows the user to specify configurations
+import enum
 import string
+from typing import Optional
 
 JOB_ID_VALID_CHARACTERS = set(string.ascii_letters + string.digits + "-_.")
 
@@ -17,6 +19,34 @@ MEADOWRUN_IO_MOUNT_LINUX = "/meadowrun/io"
 # names of resources and default values
 MEMORY_GB = "memory_gb"
 LOGICAL_CPU = "logical_cpu"
+GPU = "gpu"
+GPU_MEMORY = "gpu_memory"
+# in the resources, we set 100 - interruption probability as a non-consumable resource
+INTERRUPTION_PROBABILITY_INVERSE = "interruption_probability_inverse"
+INTEL = "intel"
+AMD = "amd"
+AVX = "avx"
+INTEL_DEEP_LEARNING_BOOST = "intel_deep_learning_boost"
+
+
+class AvxVersion(enum.IntEnum):
+    NONE = 0
+    AVX = 1
+    AVX2 = 2
+    AVX512 = 3
+
+
+def avx_from_string(s: str) -> Optional[AvxVersion]:
+    s = s.upper()
+    if s != "NONE":
+        try:
+            return AvxVersion[s]
+        except Exception:
+            pass
+
+    return None
+
+
 # these are totally arbitrary values...
 DEFAULT_MEMORY_GB_REQUIRED = 2
 DEFAULT_LOGICAL_CPU_REQUIRED = 1
