@@ -7,7 +7,17 @@ import shutil
 import tempfile
 import urllib.parse
 import zipfile
-from typing import Any, Callable, Coroutine, List, Optional, Sequence, Tuple, Union
+from typing import (
+    Any,
+    Callable,
+    Coroutine,
+    List,
+    Mapping,
+    Optional,
+    Sequence,
+    Tuple,
+    Union,
+)
 
 import filelock
 
@@ -251,14 +261,7 @@ async def _get_zip_file_code_paths(
         ]
 
     async def download_and_unzip(
-        download: Callable[
-            [
-                str,
-                str,
-                str,
-            ],
-            Coroutine[Any, Any, None],
-        ]
+        download: Callable[[str, str, str], Coroutine[Any, Any, None]],
     ) -> List[str]:
         bucket_name = decoded_url.netloc
         object_name = decoded_url.path.lstrip("/")
@@ -294,7 +297,7 @@ async def compile_environment_spec_to_container(
     TODO we should also consider creating conda environments locally rather than always
     making a container for them, that might be more efficient.
     """
-    path_to_spec, spec_hash = _get_path_and_hash(
+    path_to_spec, spec_hash, has_git_dependency = _get_path_and_hash(
         environment_spec, interpreter_spec_path
     )
 
@@ -340,30 +343,67 @@ async def compile_environment_spec_to_container(
     # the image doesn't exist locally or in ECR, so build it ourselves and cache it in
     # ECR
 
+    print(f"Building python environment in container {spec_hash}")
+
     build_args = {}
     files_to_copy = []
+    apt_packages = [
+        p for p in environment_spec.additional_software.keys() if p != "cuda"
+    ]
+    if has_git_dependency:
+        apt_packages.append("git")
+    if apt_packages:
+        build_args["APT_PACKAGES"] = " ".join(apt_packages)
+
     if environment_spec.environment_type == EnvironmentType.CONDA:
-        docker_file_name = "CondaDockerfile"
+        if apt_packages:
+            docker_file_name = "CondaDockerfile"
+        else:
+            docker_file_name = "CondaAptDockerfile"
         spec_filename = os.path.basename(path_to_spec)
         build_args["ENV_FILE"] = spec_filename
         files_to_copy.append((path_to_spec, spec_filename))
-    elif environment_spec.environment_type == EnvironmentType.PIP:
-        docker_file_name = "PipDockerfile"
-        spec_filename = os.path.basename(path_to_spec)
-        build_args["ENV_FILE"] = spec_filename
-        files_to_copy.append((path_to_spec, spec_filename))
-        build_args["PYTHON_VERSION"] = environment_spec.python_version
-    elif environment_spec.environment_type == EnvironmentType.POETRY:
-        docker_file_name = "PoetryDockerfile"
-        build_args["PYTHON_VERSION"] = environment_spec.python_version
-        files_to_copy.append(
-            (os.path.join(path_to_spec, "pyproject.toml"), "pyproject.toml")
-        )
-        files_to_copy.append((os.path.join(path_to_spec, "poetry.lock"), "poetry.lock"))
+        if "cuda" in environment_spec.additional_software:
+            build_args[
+                "CONDA_IMAGE"
+            ] = "meadowrun/cuda-conda:cuda11.6.2-cudnn8-ubuntu20.04"
+        else:
+            build_args["CONDA_IMAGE"] = "continuumio/miniconda3"
     else:
-        raise ValueError(
-            f"Unexpected environment_type {environment_spec.environment_type}"
-        )
+        # common to pip and poetry
+        if "cuda" in environment_spec.additional_software:
+            build_args["PYTHON_IMAGE"] = (
+                f"meadowrun/cuda-python:py{environment_spec.python_version}-"
+                f"cuda11.6.2-cudnn8-ubuntu20.04"
+            )
+        else:
+            build_args[
+                "PYTHON_IMAGE"
+            ] = f"python:{environment_spec.python_version}-slim-bullseye"
+
+        if environment_spec.environment_type == EnvironmentType.PIP:
+            if apt_packages:
+                docker_file_name = "PipAptDockerfile"
+            else:
+                docker_file_name = "PipDockerfile"
+            spec_filename = os.path.basename(path_to_spec)
+            files_to_copy.append((path_to_spec, spec_filename))
+            build_args["ENV_FILE"] = spec_filename
+        elif environment_spec.environment_type == EnvironmentType.POETRY:
+            if apt_packages:
+                docker_file_name = "PoetryAptDockerfile"
+            else:
+                docker_file_name = "PoetryDockerfile"
+            files_to_copy.append(
+                (os.path.join(path_to_spec, "pyproject.toml"), "pyproject.toml")
+            )
+            files_to_copy.append(
+                (os.path.join(path_to_spec, "poetry.lock"), "poetry.lock")
+            )
+        else:
+            raise ValueError(
+                f"Unexpected environment_type {environment_spec.environment_type}"
+            )
 
     docker_file_path = os.path.join(
         os.path.dirname(__file__), "docker_files", docker_file_name
@@ -378,6 +418,7 @@ async def compile_environment_spec_to_container(
     # different machines) building the identical image at the same time.
     if helper.should_push:
         try:
+            print("Caching container")
             await push_image(helper.image_name, helper.username_password)
         except aiodocker_exceptions.DockerError as e:
             print(
@@ -388,14 +429,27 @@ async def compile_environment_spec_to_container(
     return result
 
 
+def _hash_spec(spec_contents: bytes) -> str:
+    # TODO better to exclude the name and prefix for conda yml files as those
+    # are ignored
+    return hashlib.blake2b(spec_contents, digest_size=64).hexdigest()
+
+
+def _format_additional_software(additional_software: Mapping[str, str]) -> str:
+    return "\n\nmeadowrun-additional-software: " + "; ".join(
+        f"{key}: {value}" for key, value in sorted(additional_software.items())
+    )
+
+
 def _get_path_and_hash(
     environment_spec: Union[EnvironmentSpecInCode, EnvironmentSpec],
     interpreter_spec_path: str,
-) -> Tuple[str, str]:
-    def hash_spec(it: bytes) -> str:
-        # TODO probably better to exclude the name and prefix in the file as those
-        # are ignored
-        return hashlib.blake2b(it, digest_size=64).hexdigest()
+) -> Tuple[str, str, bool]:
+    """
+    Returns path_to_spec, spec_hash, has_git_dependency. spec_hash is a hash that
+    uniquely identifies the environment. path_to_spec is used slightly differently for
+    each environment type's Dockerfile. has_git_dependency is also specific to each
+    """
 
     if isinstance(environment_spec, EnvironmentSpecInCode):
         # in this case, interpreter_spec_path is a path to the root of the place where
@@ -410,15 +464,32 @@ def _get_path_and_hash(
         else:
             file_to_hash = path_to_spec
         with open(file_to_hash, "rb") as spec:
-            spec_hash = hash_spec(spec.read())
-        return path_to_spec, spec_hash
+            spec_contents_bytes = spec.read()
+        return (
+            path_to_spec,
+            _hash_spec(
+                spec_contents_bytes
+                + _format_additional_software(
+                    environment_spec.additional_software
+                ).encode("utf-8")
+            ),
+            _has_git_dependency(
+                spec_contents_bytes.decode("utf-8"), environment_spec.environment_type
+            ),
+        )
     elif isinstance(environment_spec, EnvironmentSpec):
         # in this case, interpreter_spec_path is a path to where we save the spec for
         # later processing.
         if environment_spec.environment_type == EnvironmentType.POETRY:
             # for poetry, path_to_spec will be a folder that contains pyproject.toml and
             # poetry.lock
-            spec_hash = hash_spec(environment_spec.spec_lock.encode("UTF-8"))
+            spec_contents_str = environment_spec.spec_lock
+            spec_hash = _hash_spec(
+                (
+                    spec_contents_str
+                    + _format_additional_software(environment_spec.additional_software)
+                ).encode("UTF-8")
+            )
             path_to_spec = os.path.join(interpreter_spec_path, spec_hash)
             os.makedirs(path_to_spec, exist_ok=True)
             with open(
@@ -430,7 +501,13 @@ def _get_path_and_hash(
             ) as lock_file:
                 lock_file.write(environment_spec.spec_lock)
         else:
-            spec_hash = hash_spec(environment_spec.spec.encode("UTF-8"))
+            spec_contents_str = environment_spec.spec
+            spec_hash = _hash_spec(
+                (
+                    spec_contents_str
+                    + _format_additional_software(environment_spec.additional_software)
+                ).encode("UTF-8")
+            )
 
             if environment_spec.environment_type == EnvironmentType.CONDA:
                 spec_filename = f"conda_env_spec_{spec_hash}.yml"
@@ -445,8 +522,27 @@ def _get_path_and_hash(
             with open(path_to_spec, "w", encoding="utf-8") as env_spec:
                 env_spec.write(environment_spec.spec)
 
-        return path_to_spec, spec_hash
+        return (
+            path_to_spec,
+            spec_hash,
+            _has_git_dependency(spec_contents_str, environment_spec.environment_type),
+        )
     else:
         raise ValueError(
             f"Unexpected type of environment_spec {type(environment_spec)}"
         )
+
+
+def _has_git_dependency(
+    spec_contents: str, environment_type: EnvironmentType.ValueType
+) -> bool:
+    # We just operate on bytes rather than decoding the string
+    if environment_type == EnvironmentType.PIP:
+        return any(line.startswith("git+") for line in spec_contents.splitlines())
+    elif environment_type == EnvironmentType.POETRY:
+        return any(
+            line.startswith('type = "git"') for line in spec_contents.splitlines()
+        )
+    else:
+        # TODO add support for conda
+        return False
