@@ -30,10 +30,14 @@ import json
 import tarfile
 import urllib.parse
 import urllib.request
-from typing import Tuple, Optional, List, Dict, Iterable
+from typing import TYPE_CHECKING, Tuple, Optional, List, Dict, Iterable, Any
 
 from meadowrun._vendor import aiodocker
-from meadowrun._vendor.aiodocker import containers as aiodocker_containers
+
+if TYPE_CHECKING:
+    from meadowrun._vendor.aiodocker import containers as aiodocker_containers
+from meadowrun._vendor.aiodocker import exceptions as aiodocker_exceptions
+
 import aiohttp
 
 import meadowrun.credentials
@@ -392,7 +396,10 @@ async def run_container(
     image: str,
     cmd: List[str],
     environment_variables: Dict[str, str],
+    working_directory: str,
     binds: List[Tuple[str, str]],
+    ports: List[str],
+    all_gpus: bool,
 ) -> Tuple[aiodocker_containers.DockerContainer, aiodocker.Docker]:
     """
     Runs a docker container. Examples of parameters:
@@ -419,39 +426,85 @@ async def run_container(
     # Now actually run the container. For documentation on the config object:
     # https://docs.docker.com/engine/api/v1.41/#operation/ContainerCreate
     client = await aiodocker.Docker().__aenter__()
-    container = await client.containers.run(
-        {
-            "Image": image,
-            "Cmd": cmd,
-            "Env": [f"{key}={value}" for key, value in environment_variables.items()],
-            "HostConfig": {
-                "Binds": [
-                    f"{path_on_host}:{path_in_container}"
-                    for path_on_host, path_in_container in binds
-                ],
-                # Docker for Windows and Mac automatically enable containers to access
-                # the host with host.docker.internal, but we need to add this flag for
-                # Linux machines to use this (not supported in Linux at all before
-                # Docker v20.10). This is mostly for tests, so we include it and it
-                # shouldn't cause problems even if it doesn't work. See more at
-                # agent._prepare_py_grid's discussion of replacing localhost for the
-                # coordinator address. Also:
-                # https://stackoverflow.com/questions/31324981/how-to-access-host-port-from-docker-container/43541732#43541732
-                "ExtraHosts": ["host.docker.internal:host-gateway"],
-            },
+
+    config: Dict[str, Any] = {
+        "Image": image,
+        "Cmd": cmd,
+        "Env": [f"{key}={value}" for key, value in environment_variables.items()],
+        "WorkingDir": working_directory,
+        "HostConfig": {
+            "Binds": [
+                f"{path_on_host}:{path_in_container}"
+                for path_on_host, path_in_container in binds
+            ],
+            # Docker for Windows and Mac automatically enable containers to access the
+            # host with host.docker.internal, but we need to add this flag for Linux
+            # machines to use this (not supported in Linux at all before Docker v20.10).
+            # This is mostly for tests, so we include it and it shouldn't cause problems
+            # even if it doesn't work. See more at agent._prepare_py_grid's discussion
+            # of replacing localhost for the coordinator address. Also:
+            # https://stackoverflow.com/questions/31324981/how-to-access-host-port-from-docker-container/43541732#43541732
+            "ExtraHosts": ["host.docker.internal:host-gateway"],
+            # Ideally we would have some sort of "AutoRemove after 5 minutes". With
+            # AutoRemove turned on, containers can get deleted before we are able to
+            # read off their exit codes.
+            # "AutoRemove": True,
+        },
+    }
+
+    # unfortunately PortBindings doesn't support port ranges so we expand them manually
+    # here
+    expanded_ports = []
+    for port in ports:
+        if "-" in port:
+            split = port.split("-")
+            if len(split) != 2:
+                raise ValueError(f"Bad port specification {port}")
+            for i in range(int(split[0]), int(split[1]) + 1):
+                expanded_ports.append(str(i))
+        else:
+            expanded_ports.append(port)
+
+    if expanded_ports:
+        config["HostConfig"]["PortBindings"] = {
+            f"{port}/tcp": [{"HostPort": port}] for port in expanded_ports
         }
-    )
+        config["ExposedPorts"] = {f"{port}/tcp": {} for port in expanded_ports}
+
+    if all_gpus:
+        # it would be convenient if we could just always set this, but setting this when
+        # there are no GPUs will cause the container creation to fail
+        config["HostConfig"]["DeviceRequests"] = [
+            {"Count": -1, "Capabilities": [["gpu"]]}
+        ]
+
+    container = await client.containers.run(config)
 
     return container, client
 
 
-async def get_image_environment_variables(image: str) -> Optional[List[str]]:
+async def remove_container(container: aiodocker_containers.DockerContainer) -> None:
+    # https://docs.docker.com/engine/api/v1.41/#tag/Container/operation/ContainerDelete
+    try:
+        await container.delete(v=True, force=True)
+    except aiodocker_exceptions.DockerError as e:
+        # 404 means the container doesn't exist, which is fine, as we're trying to
+        # delete it anyways
+        if e.status != 404:
+            raise
+
+
+async def get_image_environment_variables_and_working_dir(
+    image: str,
+) -> Tuple[Optional[List[str]], str]:
     """
-    Returns a list of strings like ["PATH=/foo/bar", "PYTHON_VERSION=3.9.7"] for the
-    image that we have locally.
+    Returns environment variables and working directory as set in the specified image.
+    The image must be available locally. Environment variables will be a list of strings
+    like ["PATH=/foo/bar", "PYTHON_VERSION=3.9.7"]. working_dir will be a string.
     """
     async with aiodocker.Docker() as client:
-        return (await client.images.inspect(image))["ContainerConfig"]["Env"]
+        container_config = (await client.images.inspect(image))["ContainerConfig"]
+        return container_config["Env"], container_config["WorkingDir"]
 
 
 async def delete_image(image: str) -> None:
