@@ -1,8 +1,11 @@
+from __future__ import annotations
+
 import argparse
 import asyncio
 import datetime
 import os.path
 import time
+from typing import TYPE_CHECKING
 
 import meadowrun.aws_integration.aws_install_uninstall as aws
 import meadowrun.azure_integration.azure_install_uninstall as azure
@@ -12,6 +15,7 @@ from meadowrun.aws_integration.aws_permissions_install import (
     grant_permission_to_secret,
 )
 from meadowrun.aws_integration.ec2_instance_allocation import SSH_USER
+from meadowrun.aws_integration.ec2_pricing import clear_prices_cache
 from meadowrun.aws_integration.ec2_ssh_keys import (
     download_ssh_key as ec2_download_ssh_key,
 )
@@ -47,7 +51,9 @@ from meadowrun.azure_integration.mgmt_functions.vm_adjust import (
     _deregister_and_terminate_vms,
     terminate_all_vms,
 )
-from meadowrun.run_job_core import CloudProviderType
+
+if TYPE_CHECKING:
+    from meadowrun.run_job_core import CloudProviderType
 
 
 def _strtobool(val: str) -> int:
@@ -81,6 +87,7 @@ async def async_main(cloud_provider: CloudProviderType) -> None:
         vm_instances = "EC2 instances"
         ssh_user = SSH_USER
         vm_instance_address = "<ec2-instance-public-address>"
+        ecr = "ECR"
     elif cloud_provider == "AzureVM":
         functions = "Azure Functions"
         cloud_name = "Azure"
@@ -89,6 +96,7 @@ async def async_main(cloud_provider: CloudProviderType) -> None:
         vm_instances = "Azure VMs"
         ssh_user = _MEADOWRUN_USERNAME
         vm_instance_address = "<azure-vm-public-ip>"
+        ecr = "ACR"
     else:
         raise ValueError(f"Unexpected value for cloud_provider {cloud_provider}")
 
@@ -197,6 +205,7 @@ async def async_main(cloud_provider: CloudProviderType) -> None:
             if args.clean_active:
                 aws.terminate_all_instances(region_name, False)
             _deregister_and_terminate_instances(region_name, datetime.timedelta.min)
+            clear_prices_cache()
         elif cloud_provider == "AzureVM":
             resource_group_path = await ensure_meadowrun_resource_group(region_name)
             storage_account = await ensure_meadowrun_storage_account(
@@ -236,7 +245,7 @@ async def async_main(cloud_provider: CloudProviderType) -> None:
         )
         t0 = time.perf_counter()
 
-        print("Deleting unused meadowrun-generated ECR images")
+        print(f"Deleting unused meadowrun-generated {ecr} images")
         if cloud_provider == "EC2":
             aws_delete_unused_images(region_name)
         elif cloud_provider == "AzureVM":
@@ -245,7 +254,7 @@ async def async_main(cloud_provider: CloudProviderType) -> None:
         else:
             raise ValueError(f"Unexpected cloud_provider {cloud_provider}")
         print(
-            "Deleted unused meadowrun-generated ECR images in "
+            f"Deleted unused meadowrun-generated {ecr} images in "
             f"{time.perf_counter() - t0:.2f} seconds"
         )
     elif args.command == "grant-permission-to-secret":
