@@ -14,12 +14,14 @@ import urllib.parse
 import uuid
 from enum import Enum
 from typing import (
+    TYPE_CHECKING,
     Any,
     Callable,
     Coroutine,
     Dict,
     Iterable,
     List,
+    Mapping,
     Optional,
     Sequence,
     Tuple,
@@ -48,13 +50,16 @@ from meadowrun.azure_integration.mgmt_functions.azure_core.azure_rest_api import
 from meadowrun.conda import env_export, try_get_current_conda_env
 from meadowrun.config import JOB_ID_VALID_CHARACTERS, MEADOWRUN_INTERPRETER
 from meadowrun.credentials import CredentialsSourceForService, CredentialsService
-from meadowrun.deployment import (
-    CodeDeployment,
-    InterpreterDeployment,
-    VersionedCodeDeployment,
-    VersionedInterpreterDeployment,
-)
+
+if TYPE_CHECKING:
+    from meadowrun.deployment import (
+        CodeDeployment,
+        InterpreterDeployment,
+        VersionedCodeDeployment,
+        VersionedInterpreterDeployment,
+    )
 from meadowrun.docker_controller import get_registry_domain
+from meadowrun.instance_selection import Resources
 from meadowrun.meadowrun_pb2 import (
     AwsSecretProto,
     AzureSecretProto,
@@ -246,6 +251,18 @@ class ContainerAtDigestInterpreter(ContainerInterpreterBase):
         return self.username_password_secret
 
 
+def _additional_software_to_dict(
+    additional_software: Union[Sequence[str], str, None],
+) -> Mapping[str, str]:
+    if additional_software is None:
+        return {}
+
+    if isinstance(additional_software, str):
+        return {additional_software: ""}
+
+    return {package: "" for package in additional_software}
+
+
 class InterpreterSpecFile(abc.ABC):
     """
     An abstract base class for specifying a Conda environment.yml file or a pip
@@ -264,9 +281,12 @@ class CondaEnvironmentYmlFile(InterpreterSpecFile):
         path_to_yml_file: In the context of mirror_local, this is the path to a file on
             the local disk. In the context of git_repo, this is a path to a file in the
             git repo
+        additional_software: apt packages that need to be installed to make the
+            environment work
     """
 
     path_to_yml_file: str
+    additional_software: Union[Sequence[str], str, None] = None
 
 
 @dataclasses.dataclass
@@ -281,10 +301,13 @@ class PipRequirementsFile(InterpreterSpecFile):
         python_version: A python version like "3.9" or "3.9.5". The version must be
             available on docker: https://hub.docker.com/_/python as
             python-<python_version>-slim-bullseye.
+        additional_software: apt packages that need to be installed to make the
+            environment work
     """
 
     path_to_requirements_file: str
     python_version: str
+    additional_software: Union[Sequence[str], str, None] = None
 
 
 @dataclasses.dataclass
@@ -300,11 +323,14 @@ class PoetryProjectPath(InterpreterSpecFile):
         python_version: A python version like "3.9" or "3.9.5". The version must be
             available on docker: https://hub.docker.com/_/python as
             python-<python_version>-slim-bullseye. This python version must be
-            compatible with the requirements in pyproject.toml2
+            compatible with the requirements in pyproject.toml
+        additional_software: apt packages that need to be installed to make the
+            environment work
     """
 
     path_to_project: str
     python_version: str
+    additional_software: Union[Sequence[str], str, None] = None
 
 
 class LocalInterpreter(abc.ABC):
@@ -317,6 +343,67 @@ class LocalInterpreter(abc.ABC):
         pass
 
 
+@dataclasses.dataclass
+class LocalCurrentInterpreter(LocalInterpreter):
+    """
+    Specifies the current python interpreter.
+
+    Attributes:
+        additional_software: apt packages that need to be installed to make the
+            environment work
+    """
+
+    additional_software: Union[Sequence[str], str, None] = None
+
+    async def get_interpreter_spec(self) -> InterpreterDeployment:
+        # first, check if we are in a conda environment
+        conda_env_spec = await try_get_current_conda_env()
+        if conda_env_spec is not None:
+            if platform.system() != "Linux":
+                raise ValueError(
+                    "mirror_local from a Conda environment only works from Linux "
+                    "because conda environments are not cross-platform."
+                )
+            print("Mirroring current conda environment")
+            return EnvironmentSpec(
+                environment_type=EnvironmentType.CONDA,
+                spec=conda_env_spec,
+                additional_software=_additional_software_to_dict(
+                    self.additional_software
+                ),
+            )
+
+        # next, check if we're in a poetry environment. Unfortunately it doesn't seem
+        # like there's a way to detect that we're in a poetry environment unless the
+        # current working directory contains the pyproject.toml and poetry.lock files.
+        # If for some reason this isn't the case, we'll fall through to the pip-based
+        # case, which will mostly work for poetry environments.
+        if os.path.isfile("pyproject.toml") and os.path.isfile("poetry.lock"):
+            with open("pyproject.toml", encoding="utf-8") as project_file:
+                project_file_contents = project_file.read()
+            with open("poetry.lock", encoding="utf-8") as lock_file:
+                lock_file_contents = lock_file.read()
+            print("Mirroring current poetry environment")
+            return EnvironmentSpec(
+                environment_type=EnvironmentType.POETRY,
+                spec=project_file_contents,
+                spec_lock=lock_file_contents,
+                python_version=f"{sys.version_info.major}.{sys.version_info.minor}",
+                additional_software=_additional_software_to_dict(
+                    self.additional_software
+                ),
+            )
+
+        # if not, assume this is a pip-based environment
+        print("Mirroring current pip environment")
+        return EnvironmentSpec(
+            environment_type=EnvironmentType.PIP,
+            spec=await pip_freeze_without_local_current_interpreter(),
+            python_version=f"{sys.version_info.major}.{sys.version_info.minor}",
+            additional_software=_additional_software_to_dict(self.additional_software),
+        )
+
+
 @dataclasses.dataclass
 class LocalCondaInterpreter(LocalInterpreter):
     """
@@ -327,9 +414,12 @@ class LocalCondaInterpreter(LocalInterpreter):
             `my_env_name`) or the full path to the folder of a conda environment (e.g.
             `/home/user/miniconda3/envs/my_env_name`). Will be passed to conda env
             export
+        additional_software: apt packages that need to be installed to make the
+            environment work
     """
 
     environment_name_or_path: str
+    additional_software: Union[Sequence[str], str, None] = None
 
     async def get_interpreter_spec(self) -> InterpreterDeployment:
         if platform.system() != "Linux":
@@ -340,6 +430,7 @@ class LocalCondaInterpreter(LocalInterpreter):
         return EnvironmentSpec(
             environment_type=EnvironmentType.CONDA,
             spec=await env_export(self.environment_name_or_path),
+            additional_software=_additional_software_to_dict(self.additional_software),
         )
 
 
@@ -355,10 +446,13 @@ class LocalPipInterpreter(LocalInterpreter):
         python_version: A python version like "3.9" or "3.9.5". The version must be
             available on docker: https://hub.docker.com/_/python as
             python-<python_version>-slim-bullseye.
+        additional_software: apt packages that need to be installed to make the
+            environment work
     """
 
     path_to_interpreter: str
     python_version: str
+    additional_software: Union[Sequence[str], str, None] = None
 
     async def get_interpreter_spec(self) -> InterpreterDeployment:
         return EnvironmentSpec(
@@ -368,50 +462,10 @@ class LocalPipInterpreter(LocalInterpreter):
                 self.path_to_interpreter
             ),
             python_version=self.python_version,
+            additional_software=_additional_software_to_dict(self.additional_software),
         )
 
 
-async def _get_current_local_interpreter() -> InterpreterDeployment:
-    # first, check if we are in a conda environment
-    conda_env_spec = await try_get_current_conda_env()
-    if conda_env_spec is not None:
-        if platform.system() != "Linux":
-            raise ValueError(
-                "mirror_local from a Conda environment only works from Linux because "
-                "conda environments are not cross-platform."
-            )
-        print("Mirroring current conda environment")
-        return EnvironmentSpec(
-            environment_type=EnvironmentType.CONDA, spec=conda_env_spec
-        )
-
-    # next, check if we're in a poetry environment. Unfortunately it doesn't seem like
-    # there's a way to detect that we're in a poetry environment unless the current
-    # working directory contains the pyproject.toml and poetry.lock files. If for smoe
-    # reason this isn't the case, we'll fall through to the pip-based case, which will
-    # mostly work for poetry environments.
-    if os.path.isfile("pyproject.toml") and os.path.isfile("poetry.lock"):
-        with open("pyproject.toml", encoding="utf-8") as project_file:
-            project_file_contents = project_file.read()
-        with open("poetry.lock", encoding="utf-8") as lock_file:
-            lock_file_contents = lock_file.read()
-        print("Mirroring current poetry environment")
-        return EnvironmentSpec(
-            environment_type=EnvironmentType.POETRY,
-            spec=project_file_contents,
-            spec_lock=lock_file_contents,
-            python_version=f"{sys.version_info.major}.{sys.version_info.minor}",
-        )
-
-    # if not, assume this is a pip-based environment
-    print("Mirroring current pip environment")
-    return EnvironmentSpec(
-        environment_type=EnvironmentType.PIP,
-        spec=await pip_freeze_without_local_current_interpreter(),
-        python_version=f"{sys.version_info.major}.{sys.version_info.minor}",
-    )
-
-
 @dataclasses.dataclass(frozen=True)
 class Deployment:
     interpreter: Union[
@@ -458,14 +512,18 @@ class Deployment:
         credentials_sources = []
 
         if interpreter is None:
-            interpreter_spec: Union[
-                InterpreterDeployment, VersionedInterpreterDeployment
-            ] = await _get_current_local_interpreter()
-        elif isinstance(interpreter, CondaEnvironmentYmlFile):
+            interpreter = LocalCurrentInterpreter()
+
+        if isinstance(interpreter, CondaEnvironmentYmlFile):
             with open(interpreter.path_to_yml_file, encoding="utf-8") as f:
-                interpreter_spec = EnvironmentSpec(
+                interpreter_spec: Union[
+                    InterpreterDeployment, VersionedInterpreterDeployment
+                ] = EnvironmentSpec(
                     environment_type=EnvironmentType.CONDA,
                     spec=f.read(),
+                    additional_software=_additional_software_to_dict(
+                        interpreter.additional_software
+                    ),
                 )
         elif isinstance(interpreter, PipRequirementsFile):
             with open(interpreter.path_to_requirements_file, encoding="utf-8") as f:
@@ -473,6 +531,9 @@ class Deployment:
                     environment_type=EnvironmentType.PIP,
                     spec=f.read(),
                     python_version=interpreter.python_version,
+                    additional_software=_additional_software_to_dict(
+                        interpreter.additional_software
+                    ),
                 )
         elif isinstance(interpreter, PoetryProjectPath):
             with open(
@@ -492,6 +553,9 @@ class Deployment:
                 spec=spec,
                 spec_lock=spec_lock,
                 python_version=interpreter.python_version,
+                additional_software=_additional_software_to_dict(
+                    interpreter.additional_software
+                ),
             )
         elif isinstance(interpreter, LocalInterpreter):
             interpreter_spec = await interpreter.get_interpreter_spec()
@@ -593,18 +657,27 @@ class Deployment:
             interpreter_spec = EnvironmentSpecInCode(
                 environment_type=EnvironmentType.CONDA,
                 path_to_spec=interpreter.path_to_yml_file,
+                additional_software=_additional_software_to_dict(
+                    interpreter.additional_software
+                ),
             )
         elif isinstance(interpreter, PipRequirementsFile):
             interpreter_spec = EnvironmentSpecInCode(
                 environment_type=EnvironmentType.PIP,
                 path_to_spec=interpreter.path_to_requirements_file,
                 python_version=interpreter.python_version,
+                additional_software=_additional_software_to_dict(
+                    interpreter.additional_software
+                ),
             )
         elif isinstance(interpreter, PoetryProjectPath):
             interpreter_spec = EnvironmentSpecInCode(
                 environment_type=EnvironmentType.POETRY,
                 path_to_spec=interpreter.path_to_project,
                 python_version=interpreter.python_version,
+                additional_software=_additional_software_to_dict(
+                    interpreter.additional_software
+                ),
             )
         elif isinstance(interpreter, ContainerInterpreterBase):
             interpreter_spec = interpreter.get_interpreter_spec()
@@ -755,6 +828,14 @@ async def _prepare_code_deployment(
         raise ValueError(f"Unexpected value for target {target}")
 
 
+def _needs_cuda(flags_required: Union[Iterable[str], str, None]) -> bool:
+    if flags_required is None:
+        return False
+    if isinstance(flags_required, str):
+        return "nvidia" == flags_required
+    return "nvidia" in flags_required
+
+
 @dataclasses.dataclass(frozen=True)
 class AllocCloudInstance(Host):
     """
@@ -768,31 +849,45 @@ class AllocCloudInstance(Host):
             interruption probability less than 80% can be used. Use `0` to indicate that
             only on-demand instance are acceptable (i.e. do not use spot instances)
         cloud_provider: `EC2` or `AzureVM`
+        gpus_required:
+        gpu_memory_required: Total GPU memory required across all GPUs
+        flags_required: E.g. "intel", "avx512", etc.
         region_name:
     """
 
-    logical_cpu_required: int
+    logical_cpu_required: float
     memory_gb_required: float
     interruption_probability_threshold: float
     cloud_provider: CloudProviderType
+    gpus_required: Optional[float] = None
+    gpu_memory_required: Optional[float] = None
+    flags_required: Union[Iterable[str], str, None] = None
     region_name: Optional[str] = None
 
+    def uses_gpu(self) -> bool:
+        return self.gpus_required is not None or self.gpu_memory_required is not None
+
+    def needs_cuda(self) -> bool:
+        return self.uses_gpu() and _needs_cuda(self.flags_required)
+
     async def run_job(self, job: Job) -> JobCompletion[Any]:
+        resources_required = Resources.from_cpu_and_memory(
+            self.logical_cpu_required,
+            self.memory_gb_required,
+            self.interruption_probability_threshold,
+            self.gpus_required,
+            self.gpu_memory_required,
+            self.flags_required,
+        )
         if self.cloud_provider == "EC2":
             return await run_job_ec2_instance_registrar(
                 job,
-                self.logical_cpu_required,
-                self.memory_gb_required,
-                self.interruption_probability_threshold,
+                resources_required,
                 self.region_name,
             )
         elif self.cloud_provider == "AzureVM":
             return await run_job_azure_vm_instance_registrar(
-                job,
-                self.logical_cpu_required,
-                self.memory_gb_required,
-                self.interruption_probability_threshold,
-                self.region_name,
+                job, resources_required, self.region_name
             )
         else:
             raise ValueError(
@@ -813,6 +908,9 @@ class AllocCloudInstances:
         interruption_probability_threshold: See
             [AllocCloudInstance][meadowrun.AllocCloudInstance]
         cloud_provider: Either `EC2` or `AzureVM`
+        gpus_required_per_task:
+        gpu_memory_required_per_task: Total GPU memory required across all GPUs
+        flags_required: e.g. "intel", "avx512", etc.
         num_concurrent_tasks: The number of workers to launch. In the context of a
             [run_map][meadowrun.run_map] call, this can be less than or equal to the
             number of args/tasks. Will default to half the total number of tasks if set
@@ -820,10 +918,13 @@ class AllocCloudInstances:
         region_name:
     """
 
-    logical_cpu_required_per_task: int
+    logical_cpu_required_per_task: float
     memory_gb_required_per_task: float
     interruption_probability_threshold: float
     cloud_provider: CloudProviderType
+    gpus_required_per_task: Optional[float] = None
+    gpu_memory_required_per_task: Optional[float] = None
+    flags_required: Union[Iterable[str], str, None] = None
     num_concurrent_tasks: Optional[int] = None
     region_name: Optional[str] = None
 
@@ -866,12 +967,26 @@ def _get_friendly_name(function: Callable[[_T], _U]) -> str:
     return _make_valid_friendly_name(friendly_name)
 
 
+def _prepare_ports(
+    ports: Union[Iterable[str], str, Iterable[int], int, None]
+) -> Optional[Sequence[str]]:
+    if ports is None:
+        return None
+    elif isinstance(ports, int):
+        return [str(ports)]
+    elif isinstance(ports, str):
+        return [ports]
+    else:
+        return [str(p) for p in ports]
+
+
 async def run_function(
     function: Union[Callable[..., _T], str],
     host: Host,
     deployment: Optional[Deployment] = None,
     args: Optional[Sequence[Any]] = None,
     kwargs: Optional[Dict[str, Any]] = None,
+    ports: Union[Iterable[str], str, Iterable[int], int, None] = None,
 ) -> _T:
     """
     Runs function on a remote machine, specified by "host".
@@ -887,6 +1002,10 @@ async def run_function(
             environment (code and libraries) that are needed to run this function
         args: Passed to the function like `function(*args)`
         kwargs: Passed to the function like `function(**kwargs)`
+        ports: A specification of ports to make available on the machine that runs this
+            job. E.g. 8000, "8080-8089" (inclusive). Ports will be opened just for the
+            duration of this job. Be careful as other jobs could be running on the same
+            machine at the same time!
 
     Returns:
         The result of calling `function`
@@ -943,6 +1062,10 @@ async def run_function(
         environment_variables,
         credentials_sources,
     ) = _add_defaults_to_deployment(deployment)
+    if host.needs_cuda() and isinstance(
+        interpreter, (EnvironmentSpec, EnvironmentSpecInCode)
+    ):
+        interpreter.additional_software["cuda"] = ""
 
     code = await _prepare_code_deployment(
         code, _DeploymentTarget.from_single_host(host)
@@ -955,6 +1078,8 @@ async def run_function(
         result_highest_pickle_protocol=pickle.HIGHEST_PROTOCOL,
         py_function=py_function,
         credentials_sources=credentials_sources,
+        ports=_prepare_ports(ports),
+        uses_gpu=host.uses_gpu(),
     )
     _add_deployments_to_job(job, code, interpreter)
 
@@ -967,6 +1092,7 @@ async def run_command(
     host: Host,
     deployment: Optional[Deployment] = None,
     context_variables: Optional[Dict[str, Any]] = None,
+    ports: Union[Iterable[str], str, Iterable[int], int, None] = None,
 ) -> JobCompletion[None]:
     """
     Runs the specified command on a remote machine
@@ -980,6 +1106,10 @@ async def run_command(
         deployment: See [Deployment][meadowrun.Deployment]. Specifies
             the environment (code and libraries) that are needed to run this command
         context_variables: Experimental feature
+        ports: A specification of ports to make available on the machine that runs this
+            job. E.g. 8000, "8080-8089" (inclusive). Ports will be opened just for the
+            duration of this job. Be careful as other jobs could be running on the same
+            machine at the same time!
 
     Returns:
         A JobCompletion object that contains metadata about the running of the job.
@@ -999,6 +1129,11 @@ async def run_command(
         credentials_sources,
     ) = _add_defaults_to_deployment(deployment)
 
+    if host.needs_cuda() and isinstance(
+        interpreter, (EnvironmentSpec, EnvironmentSpecInCode)
+    ):
+        interpreter.additional_software["cuda"] = ""
+
     code = await _prepare_code_deployment(
         code, _DeploymentTarget.from_single_host(host)
     )
@@ -1019,6 +1154,8 @@ async def run_command(
             command_line=args, pickled_context_variables=pickled_context_variables
         ),
         credentials_sources=credentials_sources,
+        ports=_prepare_ports(ports),
+        uses_gpu=host.uses_gpu(),
     )
     _add_deployments_to_job(job, code, interpreter)
 
@@ -1030,6 +1167,7 @@ async def run_map(
     args: Sequence[_T],
     hosts: AllocCloudInstances,
     deployment: Optional[Deployment] = None,
+    ports: Union[Iterable[str], str, Iterable[int], int, None] = None,
 ) -> Sequence[_U]:
     """
     Equivalent to `map(function, args)`, but runs distributed and in parallel.
@@ -1043,6 +1181,10 @@ async def run_map(
             many workers to provision and what resources are needed for each worker.
         deployment: See [Deployment][meadowrun.Deployment]. Specifies the environment
             (code and libraries) that are needed to run the function
+        ports: A specification of ports to make available on the machines that runs
+            tasks for this job. E.g. 8000, "8080-8089" (inclusive). Ports will be opened
+            just for the duration of this job. Be careful as other jobs could be running
+            on the same machine at the same time!
 
     Returns:
         Returns the result of running `function` on each of `args`
@@ -1053,25 +1195,34 @@ async def run_map(
     else:
         num_concurrent_tasks = min(hosts.num_concurrent_tasks, len(args))
 
+    resources_required_per_task = Resources.from_cpu_and_memory(
+        hosts.logical_cpu_required_per_task,
+        hosts.memory_gb_required_per_task,
+        hosts.interruption_probability_threshold,
+        hosts.gpus_required_per_task,
+        hosts.gpu_memory_required_per_task,
+        hosts.flags_required,
+    )
+
+    prepared_ports = _prepare_ports(ports)
+
     if hosts.cloud_provider == "EC2":
         helper = await prepare_ec2_run_map(
             function,
             args,
             hosts.region_name,
-            hosts.logical_cpu_required_per_task,
-            hosts.memory_gb_required_per_task,
-            hosts.interruption_probability_threshold,
+            resources_required_per_task,
             num_concurrent_tasks,
+            prepared_ports,
         )
     elif hosts.cloud_provider == "AzureVM":
         helper = await prepare_azure_vm_run_map(
             function,
             args,
             hosts.region_name,
-            hosts.logical_cpu_required_per_task,
-            hosts.memory_gb_required_per_task,
-            hosts.interruption_probability_threshold,
+            resources_required_per_task,
             num_concurrent_tasks,
+            prepared_ports,
         )
     else:
         raise ValueError(f"Unexpected value for cloud_provider {hosts.cloud_provider}")
@@ -1085,6 +1236,17 @@ async def run_map(
         credentials_sources,
     ) = _add_defaults_to_deployment(deployment)
 
+    uses_gpu = (
+        hosts.gpus_required_per_task is not None
+        or hosts.gpu_memory_required_per_task is not None
+    )
+    if (
+        uses_gpu
+        and _needs_cuda(hosts.flags_required)
+        and isinstance(interpreter, (EnvironmentSpec, EnvironmentSpecInCode))
+    ):
+        interpreter.additional_software["cuda"] = ""
+
     code = await _prepare_code_deployment(code, _DeploymentTarget.from_hosts(hosts))
 
     pickle_protocol = _pickle_protocol_for_deployed_interpreter()
@@ -1111,6 +1273,8 @@ async def run_map(
                     ),
                 ),
                 credentials_sources=credentials_sources,
+                ports=prepared_ports,
+                uses_gpu=uses_gpu,
             )
             _add_deployments_to_job(job, code, interpreter)
 
@@ -1118,7 +1282,8 @@ async def run_map(
                 asyncio.create_task(
                     SshHost(
                         public_address,
-                        helper.fabric_kwargs,
+                        helper.ssh_username,
+                        helper.ssh_private_key,
                         (hosts.cloud_provider, helper.region_name),
                     ).run_job(job)
                 )
