@@ -12,6 +12,7 @@ import shutil
 import sys
 import traceback
 from typing import (
+    TYPE_CHECKING,
     Any,
     Coroutine,
     Dict,
@@ -22,10 +23,12 @@ from typing import (
     Tuple,
 )
 
-from typing_extensions import Literal
+if TYPE_CHECKING:
+    from typing_extensions import Literal
+
+    from meadowrun._vendor import aiodocker
+    from meadowrun._vendor.aiodocker import containers as aiodocker_containers
 
-from meadowrun._vendor import aiodocker
-from meadowrun._vendor.aiodocker import containers as aiodocker_containers
 from meadowrun.config import (
     MEADOWRUN_AGENT_PID,
     MEADOWRUN_CODE_MOUNT_LINUX,
@@ -44,8 +47,9 @@ from meadowrun.deployment_manager import (
 )
 from meadowrun.docker_controller import (
     run_container,
-    get_image_environment_variables,
+    get_image_environment_variables_and_working_dir,
     pull_image,
+    remove_container,
 )
 from meadowrun.meadowrun_pb2 import (
     Credentials,
@@ -96,6 +100,10 @@ class _JobSpecTransformed:
     # expose files on the host machine for input/output with the container.
     container_binds: List[Tuple[str, str]]
 
+    ports: List[str]
+
+    uses_gpu: bool
+
     environment_variables: Dict[str, str] = dataclasses.field(
         default_factory=lambda: {}
     )
@@ -167,6 +175,8 @@ def _prepare_py_command(
         # we need a list, not a protobuf fake list
         list(job.py_command.command_line),
         _io_file_container_binds(io_folder, io_files),
+        list(job.ports),
+        job.uses_gpu,
         environment,
     )
 
@@ -280,6 +290,8 @@ def _prepare_py_function(
             itertools.chain(io_files, io_files_for_function, io_files_for_arguments),
         )
         + [(_FUNC_WORKER_PATH, func_worker_path)],
+        list(job.ports),
+        job.uses_gpu,
     )
 
 
@@ -419,6 +431,8 @@ async def _non_container_job_continuation(
             process.pid,
             None,
         )
+    except asyncio.CancelledError:
+        raise
     except Exception as e:
         # there was an exception while trying to get the final ProcessState
         return ProcessState(
@@ -449,7 +463,11 @@ async def _launch_container_job(
     continuation.
     """
 
-    binds: List[Tuple[str, str]] = []
+    os.makedirs("/var/meadowrun/machine_cache", exist_ok=True)
+
+    binds: List[Tuple[str, str]] = [
+        ("/var/meadowrun/machine_cache", "/meadowrun/machine_cache")
+    ]
     mounted_code_paths: List[str] = []
 
     # populate binds with code paths we need
@@ -458,6 +476,19 @@ async def _launch_container_job(
         mounted_code_paths.append(mounted_code_path)
         binds.append((path_on_host, mounted_code_path))
 
+    # get the image's environment variables and working directory
+    (
+        image_environment_variables,
+        working_dir,
+    ) = await get_image_environment_variables_and_working_dir(container_image_name)
+
+    # If the container image has a meaningful working directory set, then we want to
+    # leave that as it is. But if the working directory is "/tmp/", we use that as a
+    # placeholder to mean that we don't care about the container image's working
+    # directory. In that case, we set it to the first mounted code path.
+    if working_dir in ("/tmp/", "/tmp") and len(mounted_code_paths) > 0:
+        working_dir = mounted_code_paths[0]
+
     # If we've exposed any code paths, add them to PYTHONPATH. The normal behavior for
     # environment variables is that if they're specified in job_spec_transformed, those
     # override (rather than append to) what's defined in the container. If they aren't
@@ -471,9 +502,6 @@ async def _launch_container_job(
                 job_spec_transformed.environment_variables["PYTHONPATH"]
             ]
         else:
-            image_environment_variables = await get_image_environment_variables(
-                container_image_name
-            )
             if image_environment_variables:
                 for image_env_var in image_environment_variables:
                     if image_env_var.startswith("PYTHONPATH="):
@@ -497,7 +525,8 @@ async def _launch_container_job(
         f"{' '.join(job_spec_transformed.command_line)}; "
         f"container image={container_image_name}; "
         f"PYTHONPATH={job_spec_transformed.environment_variables.get('PYTHONPATH')} "
-        f"log_file_name={log_file_name}; code paths={','.join(code_paths)}"
+        f"log_file_name={log_file_name}; code paths={','.join(code_paths)} "
+        f"ports={','.join(port for port in job_spec_transformed.ports)}"
     )
 
     container, docker_client = await run_container(
@@ -505,7 +534,10 @@ async def _launch_container_job(
         # json serializer needs a real list, not a protobuf fake list
         job_spec_transformed.command_line,
         job_spec_transformed.environment_variables,
+        working_dir,
         binds,
+        job_spec_transformed.ports,
+        job_spec_transformed.uses_gpu,
     )
     return container.id, _container_job_continuation(
         container,
@@ -559,10 +591,8 @@ async def _container_job_continuation(
             None,
             container.id,
         )
-
-        # TODO delete the container now that we're done with it, but doesn't need to be
-        #  on the critical path
-
+    except asyncio.CancelledError:
+        raise
     except Exception as e:
         # there was an exception while trying to get the final ProcessState
         return ProcessState(
@@ -570,6 +600,10 @@ async def _container_job_continuation(
             pickled_result=pickle_exception(e, result_highest_pickle_protocol),
         )
     finally:
+        try:
+            await remove_container(container)
+        except Exception as e:
+            print(f"Warning, unable to remove container: {e}")
         await docker_client.__aexit__(None, None, None)
 
 
@@ -730,6 +764,8 @@ async def _get_credentials_for_job(
             code_deployment_credentials = await get_matching_credentials(
                 Credentials.Service.GIT, repo_url, credentials_sources
             )
+        except asyncio.CancelledError:
+            raise
         except Exception:
             # TODO ideally this would make it back to an error message for job if it
             #  eventually fails (and maybe even if it doesn't)
@@ -747,6 +783,8 @@ async def _get_credentials_for_job(
             interpreter_deployment_credentials = await get_docker_credentials(
                 repository, credentials_sources
             )
+        except asyncio.CancelledError:
+            raise
         except Exception:
             print("Error trying to turn credentials source into actual credentials")
             traceback.print_exc()
@@ -930,6 +968,8 @@ async def run_local(
             ),
             asyncio.create_task(continuation),
         )
+    except asyncio.CancelledError:
+        raise
     except Exception as e:
         # we failed to launch the process
         return (
@@ -943,6 +983,12 @@ async def run_local(
 
 @dataclasses.dataclass(frozen=True)
 class LocalHost(Host):
+    def uses_gpu(self) -> bool:
+        return False
+
+    def needs_cuda(self) -> bool:
+        return False
+
     async def run_job(self, job: Job) -> JobCompletion[Any]:
         initial_update, continuation = await run_local(job)
         if (
